/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "oirSupport.h"
#include "interpUtil.h"
#include "fitsio.h"
#include "imaging_fdata.h"

/**********************************************************
 * MIDI calibration of photometric pixel coordinates.  
 * Uses averaged, off subtracted frames from the
 * narrowband filters and a triple pinhole for the first 
 * approximation of the camera distortion. 
 * Slit and/or pinhole spectra of the foils are used for a final 
 * adjustment using the average correlation coefficient of each
 * region's image with that of the reference region.
 *    The input images are assumed to have an initial wavelength 
 * calibration.  The output file contains the fitted calibration 
 * information.
 *    The pinhole list is used for initial quesses at the distortion 
 * parameters and the final values for the reference region 
 * (region 3, 1-rel).  Frames in slitList are used in a correlation 
 * analysis to determine the best parameters.
 * The output file has the derived IMAGING_DETECTOR structure and the 
 * dedistorted averaged frames from slitList.
 * Usage:
 *    PhotomCal  ncoef infile outfile pinList slitList
 *        ncoef is the number of coefficients to fit, in order
 *                const, x, y, x*x, x*x*x, x*y
 *                4 is usually a good value
 *        infile  gives table of extra measurements in the spectra
 *                see GetInputs for details, "none" = no file.
 *        outfile is IMAGING_FDATA table with result.
 *        pinList is a blank seperated, quoted list of 
 *                averaged IMAGING_FDATA table with 
 *                narrowband filter and triple pinholes.
 *        slitList is a blank seperated, quoted list of 
 *                averaged IMAGING_FDATA table with 
 *                a foil filter and slit (or triple pinholes).
 ************************************************************* */

#define MAXREGION 12 /* Maximum number of regions */
#define MAXINPUTS 6  /* Maximum number of input spot images */
#define MAXSPOT 300  /* Maximum number of spot measurements */

/* Local prototypes   */
int CheckFiles (int nFiles, char **Files, int *ierr);
imaging_detector_table_def 
*FitSpots (char* fileName, 
	   float Flux[MAXREGION][3],
	   float Freq[MAXREGION][3], 
	   float Sky[MAXREGION][3], 
	   float SpotX[MAXREGION][3],
	   float SpotY[MAXREGION][3], int *ierr);
void FindSpots (imaging_detector_table_def *def,
		imaging_fdata_row *row,
		float Flux[MAXREGION][3],
		float Freq[MAXREGION][3], 
		float Sky[MAXREGION][3], 
		float SpotX[MAXREGION][3],
		float SpotY[MAXREGION][3]);
void doFit  (imaging_detector_table_def *def, int nFiles, int ncoef,
	     float Flux[MAXINPUTS][MAXREGION][3],
	     float Freq[MAXINPUTS][MAXREGION][3], 
	     float Sky[MAXINPUTS][MAXREGION][3], 
	     float SpotX[MAXINPUTS][MAXREGION][3],
	     float SpotY[MAXINPUTS][MAXREGION][3], 
	     char *infilename, int *ierr);
void Region3Info (float *data, int nx, int ny, 
		  float centerX[3], float centerY[3], 
		  float flux[3]);
void GetInputs (char *infile, imaging_detector_table_def *detect, 
		int iregRef, int ndata[4], float x[4][MAXSPOT], float y[4][MAXSPOT], 
		float w[4][MAXSPOT], float Sky[4][MAXSPOT], float Freq[4][MAXSPOT]);
int fitDistort1 (int ndata, int ncoef, 
		 float *dz, float *x, float *y, float *w, 
		 float *a, float *RMS);
int fitDistort2 (imaging_detector_table_def *detect, 
		 int ndata, int ncoef, int iregF, int iregR,
		 char **files, int *ierr);
void
testRegrid (int nFiles, char** Files, char *outfileName, int *ierr);
float CrossCorr (imaging_detector_table_def *detector, 
		 int ireg1, int ireg2, 
		 imaging_fdata_row *rowin1,  
		 imaging_fdata_row *rowin2);
float ObjectFun (imaging_detector_table_def *detect, 
		 int iregT, int iregR, int nrows,
		 imaging_fdata_row **rows, imaging_fdata_row *work);
void PolyFitPeak (float y1, float y2, float y3, float delta, 
		  float *xmax, float *peak);

int main (int argc, char** argv) {

   int nPFiles, nSFiles, nFiles, ncoef;
   int i, iregR, iregF, ierr;
   char outFileName[130];
   char *bang="!";
   char **pinholeList, **slitList, *fullList[100];
   char *infilename, *outfilename;
   fitsfile *outFile = NULL;
   imaging_detector_table_def *DetectorDef=NULL, *outDetectorDef=NULL;
   imaging_fdata_table_def *odata;
   imaging_fdata_row *outData = NULL;
   /* Nominal wavelength of each fitted spot */
   float Freq[MAXINPUTS][MAXREGION][3];
   /* Sky position along slit of each fitted spot */
   float Sky[MAXINPUTS][MAXREGION][3];
   /* Peak pixel position of each fitted spot */
   float SpotX[MAXINPUTS][MAXREGION][3],  SpotY[MAXINPUTS][MAXREGION][3];
   /* Peak flux in each fitted spot */
   float Flux[MAXINPUTS][MAXREGION][3];
   ierr = 0;

   /* parse inputs */
   /* How many coefficients to solve for? */
   ncoef = strtod(argv[1], NULL);

   infilename = argv[2];  /* output file name */
   outfilename = argv[3];  /* output file name */

   /* Pinhole/narrowband filter file names */
   pinholeList = oirStrSplit(argv[4], &nPFiles);

   /* Slit/foil images file names */
   slitList = oirStrSplit(argv[5], &nSFiles);

   /* Copy all to fullList */
   for (i=0; i<nPFiles; i++) fullList[i] = pinholeList[i];
   nFiles = nPFiles + nSFiles;
   for (i=nPFiles; i<nFiles; i++) fullList[i] = slitList[i-nPFiles];

   /*  How many input pinhole files?  */
   if (nPFiles>MAXINPUTS) { /* Truncate at max */
     nPFiles = MAXINPUTS;
     fprintf (stderr, "Only allowing %d input files\n", nFiles);
   }

   /* Check that input files all OK */
   if ((CheckFiles(nFiles, fullList, &ierr)) || (ierr!=0)) {
     fprintf (stderr,"Input files are incompatable\n");
     return 1;
   }

   /* tell what doing */
   fprintf (stdout," Solving for %d coefficients using narrowband/pinholes files:\n", 
	    ncoef);
   for (i=0; i<nPFiles; i++) fprintf (stdout," %s \n", pinholeList[i]);
   fprintf (stdout," Also using foil/slit/pinhole files:\n");
   for (i=0; i<nSFiles; i++) fprintf (stdout," %s \n", slitList[i]);

/*  Loop over pinhole inputs fitting spot positions */
   for (i=0; i<nPFiles; i++) {
     /* Read/fit image return imaging detector */
     DetectorDef = 
       FitSpots (pinholeList[i], Flux[i], Freq[i], Sky[i], SpotX[i], SpotY[i], &ierr);
     if (ierr != 0) {
       fprintf(stderr, "cfitsio error %i reading file %s\n", ierr, pinholeList[i]);
       fits_report_error(stderr, ierr);
       return 1;
     }
   }

   /* Make output imaging detector to be sure it's OK */
   outDetectorDef = 
     make_imaging_detector_table_def (DetectorDef->nregion, 
				      DetectorDef->ndetect,
				      DetectorDef->maxtel,
				      DetectorDef->num_dim,
				      DetectorDef->max_coef,
				      DetectorDef->instrument,
				      DetectorDef->date_obs,
				      DetectorDef->region,
				      DetectorDef->regname,
				      DetectorDef->detector,
				      DetectorDef->ports,
				      DetectorDef->correlation,
				      DetectorDef->corner,
				      DetectorDef->gain,
				      DetectorDef->naxis,
				      DetectorDef->crval,
				      DetectorDef->crpix,
				      DetectorDef->ctype,
				      DetectorDef->cd,
				      DetectorDef->dmc,
				      DetectorDef->dmp);

   /*  Do initial Fitting using spots, return results in outDetectorDef. */
   doFit (outDetectorDef, nPFiles, ncoef, Flux, Freq, Sky, SpotX, SpotY, 
	  infilename, &ierr);
   if (ierr != 0) {
     fprintf(stderr,"error %i fitting spots\n", ierr);
     return 1;
   }

   /* final fitting using all frames - adjust to align non reference with
    the reference region */
   iregR = 2; /* reference region */
   iregF = 0; /* region to fit */
   if (fitDistort2 (outDetectorDef, nSFiles, ncoef, iregF, iregR, slitList, &ierr)
       || (ierr != 0)) {
     fprintf(stderr,"error %i in final fitting region %d \n", ierr, iregF);
     return 1;
   }
   iregF = 1; /* region to fit */
   if (fitDistort2 (outDetectorDef, nSFiles, ncoef, iregF, iregR, slitList, &ierr)
       || (ierr != 0)) {
     fprintf(stderr,"error %i in final fitting region %d \n", ierr, iregF);
     return 1;
   }
   iregF = 3; /* region to fit */
   if (fitDistort2 (outDetectorDef, nSFiles, ncoef, iregF, iregR, slitList, &ierr)
       || (ierr != 0)) {
     fprintf(stderr,"error %i in final fitting region %d \n", ierr, iregF);
     return 1;
   }
   
   /* Prepare output file - add exclamation to force fitsio to write */
   strcpy(outFileName, bang);
   strcat(outFileName, outfilename);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n",outfilename, &ierr);
      fits_report_error(stderr, ierr);
      return 1;
   }
   
   odata = make_imaging_fdata_table_def (outFile, outDetectorDef, 1, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      fits_report_error(stderr, ierr);
      return 1;
   }

   /*  create and initialize dummy row */
   outData = make_imaging_fdata_row(odata);
   outData->frame = 1;

   /*   Write output and close */
   i = write_imaging_fdata_table(outFile, odata, outData, 1, &ierr);
   i = close_imaging_fdata_table(outFile, odata, &ierr);

   /* test regridding the slit files to the output */
   testRegrid (nSFiles, slitList, outfilename, &ierr);

   /*  Cleanup */
   kill_imaging_fdata_row (outData);
   /* blows up kill_imaging_fdata_table_def (odata);*/
   
   return 0;
} /* end main */

/***************************************************************
 *  Checks that the input files all have 4 regions and that
 *  they all have the same size and corner
 * Inputs:
 *    nFiles   Number of files in Files
 *    Files    List of files to check
 * Outputs:
 *    ierr     error code, 0=>OK
 * Returns
 *   0 if OK else error
 **************************************************************** */
int CheckFiles (int nFiles, char **Files, int *ierr) {
  int retCode = 1;
  imaging_detector_table_def *detect;
  int ifile, ireg, corner[MAXREGION][2], naxis[MAXREGION][2];
  fitsfile *inFile;
  imaging_fdata_table_def *idata;

  /* read data */
  for (ifile=0; ifile<nFiles; ifile++) {
    /*    Open file */
    fits_open_file(&inFile, Files[ifile], READONLY, ierr);
    if (*ierr !=0) {
      fits_report_error(stderr, *ierr);
      return retCode;
    }

    /*     Get detector section */
    detect = read_imaging_detector_table(inFile, ierr);
    if (*ierr !=0) {
      fprintf(stderr,"error %i opening detector table in %s\n", 
	     ierr, Files[ifile]);
      return retCode;
    }

    /*   Close file  */
    fits_close_file(inFile, ierr);
    if (*ierr !=0) {
      fits_report_error(stderr, *ierr);
      kill_imaging_detector_table_def (detect); /* cleanup */
      return retCode;
    }

    /* Check number of regions */
    if (detect->nregion!=4) {
      fprintf (stderr," File %s has %d regions, not 4\n",
	       Files[ifile], detect->nregion);
      kill_imaging_detector_table_def (detect); /* cleanup */
      return retCode;
    }

    /* if this is the first file save the information */
    if (ifile==0) {
      for (ireg=0; ireg<detect->nregion; ireg++) {
	corner[ireg][0] = detect->corner[ireg][0];
	corner[ireg][1]= detect->corner[ireg][1];
	naxis[ireg][0] = detect->naxis[ireg][0];
	naxis[ireg][1] = detect->naxis[ireg][1];
      }
    } else {
      /* Check that this file compatable with the first */
      for (ireg=0; ireg<detect->nregion; ireg++) {
	if ((corner[ireg][0] != detect->corner[ireg][0]) || 
	    (corner[ireg][1] != detect->corner[ireg][1]) || 
	    (naxis[ireg][0] != detect->naxis[ireg][0]) || (naxis[ireg][1] != detect->naxis[ireg][1])) {
	  fprintf (stderr," File %s region %d incompatable \n",
		   Files[ifile], ireg);
	  fprintf (stderr,"  corner (%d,%d) not (%d,%d)\n size  (%d,%d) not (%d,%d)\n",
		   corner[ireg][0],corner[ireg][1], detect->corner[ireg][0], detect->corner[ireg][1],
		   naxis[ireg][0],naxis[ireg][1], detect->naxis[ireg][0], detect->naxis[ireg][1]);
	  kill_imaging_detector_table_def (detect); /* cleanup */
	  return retCode;
	}
      } /* end loop over regions */
    } /* end of compatability check */

    kill_imaging_detector_table_def (detect); /* cleanup */
  } /* end loop over files */

  retCode = 0;
  return retCode;
 } /* end CheckFiles */

/***************************************************************
 *  Reads averaged frames from MIDI triple pinhole/narrowband 
 *  filter data and determines locations of the spots.
 * Inputs:
 *    fileName Name of averaged imaging_fdata_table to read
 * Outputs:
 *    Flux     Flux in each fitted spot per region and spot
 *    Freq     Initial Frequency (Hz) of each spot
 *    Sky      Sky position along slit of each spot
 *    SpotX    X position on detector of each spot.
 *    SpotY    Y position on detector of each spot.
 *    ierr     error code, 0=>OK
 * Returns
 *   imaging_detector_table_def for file.
 **************************************************************** */
imaging_detector_table_def 
*FitSpots (char* fileName, 
	   float Flux[MAXREGION][3],
	   float Freq[MAXREGION][3], 
	   float Sky[MAXREGION][3], 
	   float SpotX[MAXREGION][3],
	   float SpotY[MAXREGION][3], int *ierr)
{
   fitsfile *inFile;
   imaging_detector_table_def *DetectorDef = NULL;
   imaging_fdata_table_def *idata;
   imaging_fdata_row *inData=NULL;
   int i, iRow;

   if (*ierr!=0) return NULL; /* Previous Error condition? */

   /*    Open file */
   i = fits_open_file(&inFile, fileName, READONLY, ierr);
   if (*ierr !=0) {
     fprintf(stderr,"error %i opening file %s\n", *ierr, fileName);
     fits_report_error(stderr, *ierr);
     return DetectorDef ;
   }

   /*     Get detector section */
   DetectorDef = read_imaging_detector_table(inFile, ierr);
   if (*ierr !=0) {
     fprintf(stderr,"error %i opening detector table in %s\n", 
	    ierr, fileName);
     fits_report_error(stderr, *ierr);
     return NULL;
   }
   
   /*    Open table*/
   idata = open_imaging_fdata_table(inFile, ierr);
   if (*ierr !=0) {
     fprintf(stderr,"error %i opening imaging data in file %s\n", 
            *ierr, fileName);
     fits_report_error(stderr, *ierr);
     return DetectorDef;
   }

   /*    Create table row */
   inData = make_imaging_fdata_row(idata);
   
   /*    Read row */
   iRow = 1;
   i = read_imaging_fdata_table(inFile, idata, inData, iRow, ierr);
   if (*ierr !=0) {
     fprintf(stderr, "error %i reading imaging data line %i in file %s\n", 
	    *ierr, iRow, fileName);
     fits_report_error(stderr, *ierr);
     return DetectorDef;
   }

   /*   Close file and table */
   i= close_imaging_fdata_table(inFile, idata, ierr);

   /* Find Spots in row data */
   FindSpots (DetectorDef, inData, Flux, Freq, Sky, SpotX, SpotY);

   /* Cleanup */
   kill_imaging_fdata_row(inData);
   /* blows up kill_imaging_fdata_table_def (idata);*/

   return DetectorDef;
} /* end FitSpots */

/***************************************************************
 *  Fit geometry model to spot data and return results in def
 * Inputs:
 *    def      imaging_detector_table_def for initial guess and 
 *             results
 *    nFiles   Number of files read (3 spots per)
 *    ncoef    Number of coefficients to solve for
 *    Flux     Flux in each fitted spot - used as weight in fitting
 *    Freq     Initial Frequency (Hz) of each spot
 *    Sky      Sky position along slit of each spot
 *    SpotX    X position on detector of each spot.
 *    SpotY    Y position on detector of each spot.
 *    infilename name of file containing extra measurements
 * Outputs:
 *    ierr      error code, 0=>OK
 **************************************************************** */
void doFit  (imaging_detector_table_def *def, int nFiles, int ncoef,
	     float Flux[MAXINPUTS][MAXREGION][3],
	     float Freq[MAXINPUTS][MAXREGION][3], 
	     float Sky[MAXINPUTS][MAXREGION][3], 
	     float SpotX[MAXINPUTS][MAXREGION][3],
	     float SpotY[MAXINPUTS][MAXREGION][3], 
	     char *infilename, int *ierr)
{
  int i, j, iReg, ndata;
  float dx[MAXSPOT+12], dy[MAXSPOT+12], x[MAXSPOT+12], y[MAXSPOT+12], w[MAXSPOT+12];
  float RMS, coefX[MAXCOEF], coefY[MAXCOEF];
  int addData[4];
  float addX[4][MAXSPOT], addY[4][MAXSPOT], addW[4][MAXSPOT];
  float addSky[4][MAXSPOT], addFreq[4][MAXSPOT];
  if (*ierr!=0) return; /* Previous Error condition? */
  if (!def) return; /* sanity check */

  /* read any additions measurements */
  GetInputs (infilename, def, 2, addData, addX, addY, addW, addSky, addFreq);

  /* Loop over region */
  for (iReg=0; iReg<def->nregion; iReg++) {

    /* Set y increment (position along slit) 
     use region 3 (1-rel) as reference */
    def->cd[iReg][3] =  (Sky[0][2][2] - Sky[0][2][0]) / 
      (SpotY[0][2][2] - SpotY[0][2][0]);
    def->crpix[iReg][1] = SpotY[0][2][1];

    /* Convert data into form for fitting */
    ndata = 0;
    for (i=0; i<nFiles; i++) {
      /* Three spots each file */
      for (j=0; j<3; j++) {
	x[ndata] = SpotX[i][iReg][j];
	y[ndata] = SpotY[i][iReg][j];
	w[ndata] = Flux[i][iReg][j];
	/* Position offsets from nominal (true-apparent) */
	dx[ndata] = (Freq[i][iReg][j] - def->crval[iReg][0]) / def->cd[iReg][0]
	  + def->crpix[iReg][0] - x[ndata];
	dy[ndata] = (Sky[i][iReg][j] - def->crval[iReg][1]) / def->cd[iReg][3]
	  + def->crpix[iReg][1] - y[ndata];
	ndata++;
      }
    } /* end loop over files */

    /* Add any additional data */
    for (j=0; j<addData[iReg]; j++) {
      x[ndata] = addX[iReg][j];
      y[ndata] = addY[iReg][j];
      w[ndata] = addW[iReg][j];
      dx[ndata] = (addFreq[iReg][j] - def->crval[iReg][0]) / def->cd[iReg][0]
	+ def->crpix[iReg][0] - x[ndata];
      dy[ndata] = (addSky[iReg][j] - def->crval[iReg][1]) / def->cd[iReg][3]
	+ def->crpix[iReg][1] - y[ndata];
      ndata++;
   } /* end adding additional data */
    
     for (i=0; i<ndata; i++) {
      fprintf (stdout,"  data %d x %f y %f w %f dx %f dy %f\n",
	       i, x[i], y[i], w[i], dx[i], dy[i]);
    }
  /* Fit distortions */
    if (fitDistort1 (ndata, ncoef, dx, x, y, w, coefX, &RMS))
      {    
	fprintf (stderr,"doFit: fit for X distortion failed\n");
	return;
      }
    /* debug - print results */
    fprintf (stdout,"region %d X fit RMS %f coef %f %f %f %f %f\n",
	     iReg, RMS, coefX[0], coefX[1], coefX[2], coefX[3],coefX[4]);
    for (i=0; i<ndata; i++) {
      fprintf (stdout,"  resid %d x %f y %f w %f resid %f\n",
	       i, x[i], y[i], w[i], dx[i]);
    }

    if (fitDistort1 (ndata, ncoef, dy, x, y, w, coefY, &RMS))
      {    
	fprintf (stderr,"doFit: fit for Y distortion failed\n");
	return;
      }
    fprintf (stdout,"region %d Y fit RMS %f coef %f %f %f %f %f\n",
	     iReg, RMS, coefY[0], coefY[1], coefY[2], coefY[3],coefY[4]);
    for (i=0; i<ndata; i++) {
      fprintf (stdout,"  resid %d x %f y %f w %f resid %f\n",
	       i, x[i], y[i], w[i], dy[i]);
    }

    /* Set coefficients */
    minrtsSetPoly  (def, 0, iReg,  ncoef, coefX);
    minrtsSetPoly  (def, 1, iReg,  ncoef, coefY);
  } /* end loop over region */
    
} /* end doFit */

/***************************************************************
 *  Determine centroid of triple pinhole spots from MIDI narrowband
 *  filters
 * Inputs:
 *    def      imaging_detector_table_def for initial guess and 
 *             results
 *    row      imaging_fdata row to use
 * Outputs:
 *    Flux     Flux in each fitted spot - used as weight in fitting
 *    Freq     Frequency (Hz) of each spot
 *    Sky      Sky position along slit of each spot
 *    SpotX    X position on detector of each spot.
 *    SpotY    Y position on detector of each spot.
 **************************************************************** */
void FindSpots (imaging_detector_table_def *def,
		imaging_fdata_row *row,
		float Flux[MAXREGION][3],
		float Freq[MAXREGION][3], 
		float Sky[MAXREGION][3], 
		float SpotX[MAXREGION][3],
		float SpotY[MAXREGION][3])
{
   int nRegion, iReg;
   int nxList[MAXREGION], nyList[MAXREGION];
   float lineFreq;

   /* Detector region information */
   nRegion = def->nregion;
   for (iReg=0;iReg<nRegion;iReg++) {
     nxList[iReg] = def->naxis[iReg][0];
     nyList[iReg] = def->naxis[iReg][1];
   }

   /*   Get info for each region */
   for (iReg=0; iReg<nRegion; iReg++) {
       Region3Info (row->data[iReg], nxList[iReg], nyList[iReg], 
		   SpotX[iReg], SpotY[iReg], Flux[iReg]);
   }


   /* Use Region 3 (1-rel), center spot to set line frequency */
   lineFreq = def->crval[2][0] + 
	 (SpotX[2][1] - def->crpix[2][0]) * def->cd[2][0];
   fprintf (stdout,"Line frequency %g\n",lineFreq);

   for (iReg=0; iReg<nRegion; iReg++) {
     /* Set values of frequency */
     Freq[iReg][0] = lineFreq;
     Freq[iReg][1] = lineFreq;
     Freq[iReg][2] = lineFreq;
     Sky[iReg][0] = -1.0;
     Sky[iReg][1] =  0.0;
     Sky[iReg][2] =  1.0;
     if (Flux[iReg][0]<0.0) Flux[iReg][0] = 0.0;
     if (Flux[iReg][1]<0.0) Flux[iReg][1] = 0.0;
     if (Flux[iReg][2]<0.0) Flux[iReg][2] = 0.0;

     /* Print info */
     fprintf (stdout," Region %d X %f Y %f Flux %f Fq %g\n",
	      iReg, SpotX[iReg][0], SpotY[iReg][0], Flux[iReg][0], Freq[iReg][0]);
     fprintf (stdout,"          X %f Y %f Flux %f Fq %g\n",
	      SpotX[iReg][1], SpotY[iReg][1], Flux[iReg][1], Freq[iReg][0]);
     fprintf (stdout,"          X %f Y %f Flux %f Fq %g\n",
	      SpotX[iReg][2], SpotY[iReg][2], Flux[iReg][2], Freq[iReg][0]);
   }
} /* end FindSpots */

/***************************************************************
 * Determine the centroids and total fluxs for triple pinhole 
 * for each pinhole in a region from the MIDI detector.
 * Center determined from first moment.
 * Inputs:
 *    data     2-D data array
 *    nx       first (fastest) dimension of data
 *    ny       second (slowestest) dimension of data
 *    centerX "X" centroids in each spot, 1-rel
 *    centerY "Y" centroids in each spot, 1-rel
 *    flux     Summed fluxs in each spot
 **************************************************************** */

void Region3Info (float *data, int nx, int ny, 
		  float centerX[3], float centerY[3], 
		  float flux[3])
{
  int ix, iy, peak, hi, lo;
  float val, maxv, sum, sum1, sum2;
  float *rowFlux=NULL, *rowPeak=NULL;
  
  /* init results */
  flux[0]    = 0.0;
  centerX[0] = 0.0;
  centerY[0] = 0.0;
  flux[1]    = 0.0;
  centerX[1] = 0.0;
  centerY[1] = 0.0;
  flux[2]    = 0.0;
  centerX[2] = 0.0;
  centerY[2] = 0.0;
  
  /* Create work arrays */
  rowFlux = (float*)malloc(ny*sizeof(float));
  rowPeak = (float*)malloc(ny*sizeof(float));
  
  /* Reduce each spectral row to an X centroid and "peak" */
  for (iy =0; iy<ny; iy++) {
    
    /* find maximum in row */
    maxv = -1.0e20;
    peak = -1;
    for (ix = 0; ix<nx; ix++) {
      val = data[iy*nx+ix];
      if (val>maxv) {
	peak = ix;
	maxv = val;
      }
    }
    
    /* get x centroid using +/- 2 pixels and first moment */
    lo = peak - 2;
    if (lo<0) lo = 0;
    hi = peak + 2;
    if (hi>=nx) hi = nx-1;
    sum = sum1 = 0.0;
    for (ix=lo; ix<=hi; ix++) {
      val = data[iy*nx+ix];
      sum  += val;
      sum1 += val * (float)ix;
    }
    
    rowFlux[iy] = sum;
    rowPeak[iy] = sum1/sum;
    /* debug
       fprintf (stdout," %d %f %f\n",iy,rowFlux[iy],rowPeak[iy]); */
  } /* end loop compressing rows */
  
  /* find global max value in compressed peaks*/
  maxv = -1.0e9;
  /* avoid edges */
  for (iy=3; iy<ny-3; iy++) {
    if (maxv<rowFlux[iy]) maxv = rowFlux[iy];
  }

  /* Use loop of length 1 to all a break to deal with no further peaks */
  for (ix=0; ix <1; ix++) {
  
    /* Find first max, must be within 30% of global */
    peak = 0;
    for (iy=2; iy<ny-2; iy++) {
      val = rowFlux[iy];
      if ((val>0.5*maxv) && (rowFlux[iy+1]<0.95*val)) {
	/* High enough and next is decreasing, call this a max */
	peak = iy;
	break;
      }
    }
    
    /* debug
       fprintf (stdout,"Peak1 %d %f %f\n",peak,val,maxv); */
    /* get centroid using +/- 1 pixels and first moment */
    lo = peak - 1;
    if (lo<0) lo = 0;
    hi = peak + 1;
    if (hi>=ny) hi = ny-1;
    sum = sum1 = sum2 = 0.0;
    for (iy=lo; iy<=hi; iy++) {
      val = rowFlux[iy];
      sum  += val;
      sum1 += val * rowPeak[iy];
      sum2 += val * (float)iy;
    }
    /* save results */
    flux[0]    = sum;
    centerX[0] = 1.0 + sum1/sum;  /* 1-rel */
    centerY[0] = 1.0 + sum2/sum;
    
    /* Find second max, must be within 50% of global */
    if (hi>(ny-2)) break;  /* hit end? */
    peak = hi;
    for (iy=hi+3; iy<ny-2; iy++) {
      val = rowFlux[iy];
      if ((val>0.5*maxv) && (rowFlux[iy+1]<0.95*val)) {
	/* High enough and next is decreasing, call this a max */
	peak = iy;
	break;
      }
    }
    /* debug
       fprintf (stdout,"Peak2 %d %f %f\n",peak,val,maxv); */
    /* get centroid using +/- 1 pixels and first moment */
    lo = peak - 1;
    if (lo<0) lo = 0;
    hi = peak + 1;
    if (hi>=ny) hi = ny-1;
    sum = sum1 = sum2 = 0.0;
    for (iy=lo; iy<=hi; iy++) {
      val = rowFlux[iy];
      sum  += val;
      sum1 += val * rowPeak[iy];
      sum2 += val * (float)iy;
    }
    /* save results */
    flux[1]    = sum;
    centerX[1] = 1.0 + sum1/sum;
    centerY[1] = 1.0 + sum2/sum;
    
    /* Find last max, must be within 50% of global */
    peak = -1;
    if (hi>(ny-2)) break;  /* hit end? */
    for (iy=hi+3; iy<ny-2; iy++) {
      val = rowFlux[iy];
      if ((val>0.5*maxv) && (rowFlux[iy+1]<0.95*val)) {
	/* High enough and next is decreasing, call this a max */
	peak = iy;
  	break;
      }
    }
    /* debug
       fprintf (stdout,"Peak3 %d %f %f %f %d\n",peak,val,maxv,rowFlux[peak],iy); */
    /* get centroid using +/- 1 pixels and first moment */
    lo = peak - 1;
    if (lo<0) lo = 0;
    hi = peak + 1;
    if (hi>=ny) hi = ny-1;
    sum = sum1 = sum2 = 0.0;
    for (iy=lo; iy<=hi; iy++) {
       val = rowFlux[iy];
      sum  += val;
      sum1 += val * rowPeak[iy];
      sum2 += val * (float)iy;
    }
    /* save results */
    flux[2]    = sum;
    centerX[2] = 1.0 + sum1/sum;
    centerY[2] = 1.0 + sum2/sum;
  } /* end or artificial loop */
  
  /* cleanup work arrays */
  if (rowFlux) free(rowFlux);
  if (rowPeak) free(rowPeak);
  
} /* end Region3Info */

/***************************************************************
 *   Read extra measurements for fitting distortion polynomial.
 *   These are likely from MIDI lab spectra with the triple 
 *   pinhole and the CALIB1 or CALIB2 foil filters.
 *   Corresponding features can be identified and measured in 
 *   the different  regions.
 *      The measurements are read from an ascii  file with the 
 *   detector positions of a spectral/spatial feature in all regions
 *   on a single line.  Comments are lines starting with '!' in col 1.
 *   The befinning of an example follows:
 *! Measured points from CALIB2 and triple pinhole,  input to PhotomCal1
 *! All points on the same line have the same spatial and wavelength position
 *! values are pixels numbers (1-rel) on the detector, -1=> not measured
 *!Sky   region 0      region 1        region 2     region 3
 *-1.0  61.25 46.25   57.75  86.25   56.25 146.75   -1.0   -1.0
 * 0.0  60.25 52.25   57.75  92.25   56.25 152.75   -1.0   -1.0
 * 1.0  60.75 58.75   57.75  98.25   56.25 158.75   -1.0   -1.0
 * "Sky" is -1, 0, 1 for the lower, middle and upper pinholes of one 
 * of the MIDI triple pinholes.
 * Pixel locations can be determined from a frame image made by ftable2cube
 * from an averaged, background subtracted imaging_fdata.
 *   Inputs
 *     infile  Name of the input file, if "none" then ignore
 *     detect  imaging_detector describing frames
 *     iregRef Reference region number, 0-rel (used to determine frequencies)
 *   Outputs
 *     ndata  number of values per region (max MAXSPOT)
 *     x      Array of "x" locations per region, <0 -> no value
 *     y      Array of "y" locations per region, <0 -> no value
 *     w      Array of weights per region (arbitrarily set to 1000.0)
 *     Sky    Array of "sky" positions per region
 *     Freq   Array of frequencies (Hz) per region
 **************************************************************** */
void GetInputs (char *infile, imaging_detector_table_def *detect, 
		int iregRef, int ndata[4], float x[4][MAXSPOT], float y[4][MAXSPOT], 
		float w[4][MAXSPOT], float Sky[4][MAXSPOT], float Freq[4][MAXSPOT])
{
  char OK=1, line[201];
  int lineMax=200, count, ireg;
  FILE *inFile;
  float skyPos, posx[4], posy[4], skyFreq;

  /* initialize */
  ndata[0] = ndata[1] = ndata[2] = ndata[3] = 0;

  /* anything to do? */
  if (!strcmp(infile,"none")) return;

  inFile = fopen (infile, "r");
  /* loop through file */
  while (OK) {
    OK = fgets (line, lineMax, inFile)!=NULL;
    if (errno!=0) {      /* it went wrong */
      fprintf (stderr,"GetInputs: error reading input file %s\n",infile);
      fclose(inFile);
      return;
    }
    if (!OK) break;  /* done? */

    /* Comment? */
    if (line[0]=='!') continue;

    /* parse */
    count = sscanf (line, "%f %f %f %f %f %f %f %f %f ",
		    &skyPos, &posx[0], &posy[0], &posx[1], &posy[1], 
		    &posx[2], &posy[2], &posx[3], &posy[3]);
    /* Check parts count */
    if (count!=9) {
      fprintf (stderr,"GetInputs: error parsing line:\n   %s\n",line);
      fclose(inFile);
      return;
    }

    /* Correct for windowing, 1-rel*/
    for (ireg=0; ireg<4; ireg++) {
      if (posx[ireg]>0) posx[ireg] -= (float)detect->corner[ireg][0] - 1.0;
      if (posy[ireg]>0) posy[ireg] -= (float)detect->corner[ireg][1] - 1.0;
    }

    /* Get sky frequency from reference region */
    skyFreq = detect->crval[iregRef][0] + 
	 (posx[iregRef] - detect->crpix[iregRef][0]) * detect->cd[iregRef][0];

    /* are we full up? */
    if ((ndata[0]>=MAXSPOT) || (ndata[1]>=MAXSPOT) || 
	(ndata[2]>=MAXSPOT) || (ndata[3]>=MAXSPOT)) {
      fprintf (stderr,"GetInputs: exceeded maximum %d, ignoring rest\n",MAXSPOT);
      fclose(inFile);
      return;
   }

    /* add to output arrays */
    for (ireg=0; ireg<4; ireg++) {
      if (posx[ireg]>0) { /* valid measurement? */
	x[ireg][ndata[ireg]]    = posx[ireg];
	y[ireg][ndata[ireg]]    = posy[ireg];
	w[ireg][ndata[ireg]]    = 1000.0;
	Sky[ireg][ndata[ireg]]  = skyPos;
	Freq[ireg][ndata[ireg]] = skyFreq;
	ndata[ireg]++;
      }
    }
  } /* end loop through file */

  /* close */
  fclose(inFile);

} /* end GetInputs */
 
/***************************************************************
 *   Fits distortion model to a set of measured MIDI detector offsets
 *   dz_j = a[0] + a[1]*x_j + a[2]*y_j + a[3]*x_j*x_j  + a[4]*x_j*x_j*x_j +
 *          a[5]*x_j*y_j
 *   for observation j and 
 *   dz_j is the measured offset from the proper position
 *   x_j, y_j are the detector pixel locations for observation j.
 *   Solution uses a relaxation method from Fred Schwab:
 *   Pn+1 = Pn + atan2 ((dChi2/dP), (d2Chi2/dP2))
 *     for each parameter P where n or n+1 indicates a given iteration, 
 *     dChi2/dP is the first partial derivative of Chi squared wrt P,
 *     d2Chi2/d2P is the second partial derivative of Chi squared wrt P,
 *     Chi2 = Sum (w_j * (dz_j - dz_j_model)**2)
 *       dz_j_model is model value as defined above
 *   Inputs
 *      ndata   Number of observations (max 30)
 *      ncoef   Number of terms to fit in polynomial expansion (max 6)
 *      dz      Measured position offsets (pixels) per obs.
 *      x       "X" pixel positions of the observations
 *      y       "Y" pixel positions of the observations
 *      W       Weights of the observations
 *   Outputs
 *      a       Polynomial coefficients (ncoef fitted)
 *      dz      Post fit Residuals from the model
 *      RMS     Post fit RMS residual in pixels
 *   Returns 0 is succeeds, else failed.
 **************************************************************** */
int fitDistort1 (int ndata, int ncoef, 
		 float *dz, float *x, float *y, float *w, 
		 float *a, float *RMS)
{
int i, j, iter, retCode = 1;
  int convgd;
  float sum, sumWt, resid, tcoef[MAXCOEF];
  float sump1[MAXSPOT], sump2[MAXSPOT],  mpt[MAXCOEF];
  float tol, delta[MAXCOEF], wx, test, norm, c2, lastc2;
  float pd1[MAXCOEF], pd2[MAXCOEF];

  /* sanity checks */
  if (ndata<1) return retCode;
  if (ncoef<1) return retCode;
  if (!dz) return retCode;
  if (!x) return retCode;
  if (!y) return retCode;
  if (!w) return retCode;
  if (!a) return retCode;
  if (ncoef>=ndata) {
    fprintf (stderr,"fitDistort1 too little data for fit %d <= %d\n",
	     ndata, ncoef);
    return retCode;
  }
  if (ncoef>MAXCOEF) ncoef = MAXCOEF;
  if (ndata>MAXSPOT) ndata = MAXSPOT;

  /* Initialize results */
  *RMS=1.0E9;
  for (i=0; i<MAXCOEF; i++) a[i] = 0.0;

  /* Crude first guesses */
  /* first term is average */
  sum = sumWt = 0.0;
  for (i=0; i<ndata; i++) {
    sumWt += w[i];
    sum += w[i]*dz[i];
  }
  if (sumWt>0.0) a[0] = sum / sumWt;
  else return retCode;

  /* debug initial model */
  /*a[0] = 0.45;
    a[1] = 0.015;*/
  a[0] = 0.0;
  /* fprintf (stdout,"Initial model %f %f %f %f\n",a[0],a[1],a[2],a[3]);
     fprintf (stdout,"Initial data %f %f %f %f %f %f %f %f %f\n",
     dz[0], dz[1],dz[2],dz[3],dz[4],dz[5],dz[6],dz[7],dz[8]);
     fprintf (stdout, "x %f %f %f %f %f %f %f %f %f\n",
     x[0],x[1],x[2],x[3],x[4],x[5],x[6],x[7],x[8]);
     fprintf (stdout, "y %f %f %f %f %f %f %f %f %f\n",
     y[0],y[1],y[2],y[3],y[4],y[5],y[6],y[7],y[8]);
     fprintf (stdout, "w %f %f %f %f %f %f %f %f %f\n",
     w[0],w[1],w[2],w[3],w[4],w[5],w[6],w[7],w[8]);*/

  /* Iteration loop */
  lastc2 = 1.0e20;
  for (iter=1; iter<=10000; iter++) {
    convgd = 0;  /* Can decide converged later */

    for (i=0; i<ncoef; i++) {  /* initialize sums */
      sump1[i] = 0.0;
      sump2[i] = 1.0e-20;
    }

    /* Loop over data */
    for (j=0; j<ndata; j++) {
      /* model polynomial terms - calculate for all possible terms */
      SETMPT(mpt,x[j],y[j]);
      
      /* residual */
      resid = dz[j];
      for (i=0; i<ncoef; i++) resid -= a[i]*mpt[i];

      /* sum 1st, 2nd partial derivatives of Chi^2 */
      for (i=0; i<ncoef; i++) {  
	pd1[i] = -w[j] * resid * mpt[i];
	pd2[i] = +w[j] * mpt[i] * mpt[i];
	sump1[i] += pd1[i];
	sump2[i] += pd2[i];
      }
    } /* end loop over data */

    /* trial update solutions with damping to force motion towards convergence */
    wx = 2.0;
    if (iter==1) wx = 0.01;
    while (wx>1.0e-10) { /* loop 'til improvement - but not forever */
      wx *= 0.5;

      /* get gradient norm */
      norm = 0.0;
      /* Convergence criteria, lower bar each attempt */
      tol = 0.5e-6 + iter*1.0e-5;
      for (i=0; i<ncoef; i++) {
	delta[i] = atan2(sump1[i], sump2[i]);
	test = tol * fabs(a[i]);
	tcoef[i] = a[i] - wx * delta[i];  /* trial update */
	convgd = convgd && (fabs(delta[i])<test); /* convergence test */
	norm += delta[i]*delta[i];
      }

      /* Determine chi^2 */
      c2 = 0.0;
      /* Loop over data */
      for (j=0; j<ndata; j++) {
	/* model polynomial terms - calculate for all possible terms */
	SETMPT(mpt,x[j],y[j]);

	resid = dz[j]; /* residual */
	for (i=0; i<ncoef; i++) resid -= tcoef[i]*mpt[i];
	c2 += w[j] * resid * resid;
      } /* end loop over data */

      /* Force chi^2 to decrease, when they do break the trial loop */
      if (c2<lastc2) break;
    } /* end trial update loop */

    lastc2 = c2;  /* save chi squares */

    /* save updated coefficients */
    for (i=0; i<ncoef; i++) a[i] = tcoef[i];

    /* are we converged?  If so break from loop */
    if (convgd || ((iter>5) && (norm<1.0e-8))) break;
    /* debug 
       fprintf (stdout," iter %d c2 %g wx %f norm %g  %g %g %g %g %g %g\n", 
       iter, c2,wx, norm, a[0], a[1], a[2], a[3], a[4], a[5]);
       fprintf (stdout,"        %g %g %g %g %g %g\n", 
       sump1[0], sump2[0], delta[0], sump1[1], sump2[1], delta[1]);*/
  } /* end iteration loop */

  /* debug */
  fprintf (stdout," Number of iterations %d chi2 %f\n", iter-1, c2);

  /* Compute residuals */
  /* Loop over data */
  sum = 0.0;
  for (j=0; j<ndata; j++) {
    if (w[j]>0.0) {
      /* model polynomial terms - calculate for all possible terms */
      SETMPT(mpt,x[j],y[j]);
      
      resid = dz[j]; /* residual */
      for (i=0; i<ncoef; i++) resid -= tcoef[i]*mpt[i];
      dz[j] = resid;  /* replace data with residuals */
      sum += resid*resid;   /* Sum for RMS residual */
    }
  } /* end loop over data */
  *RMS = sqrt (sum/ndata); /* return RMS residual */


  retCode = 0;
  return retCode;
} /* end fitDistort1 */

/***************************************************************
 *   Fits distortion model from cross correlations of region 
 *   images to a reference region.
 *   Each cycle makes the correction (or fraction thereof) to maximize
 *   the objective function for each parameter independently
 *   Uses averaged background subtracted MIDI frames as imaging_fdata
 *   of multiple lab spectra.  These should include multiple pinholes 
 *   to constrain the spectra in the Y (spatial) direction.
 *   The distortion model is:
 *   dz_j = a[0] + a[1]*x_j + a[2]*y_j + a[3]*x_j*x_j  + a[4]*x_j*x_j*x_j +
 *          a[5]*x_j*y_j
 *   for observation j and 
 *   dz_j is the measured offset from the proper position
 *   x_j, y_j are the detector pixel locations for observation j.
 *   Initial values of the coefficients are taken from detect
 *   Inputs
 *      detect  imaging_detector_table_def for initial guess of dedistortion
 *      ndata   Number of files
 *      ncoef   Number of terms to fit in polynomial expansion (max 6)
 *      iregF   Region to fit
 *      iregR   Reference region number.
 *      files   Names of the input files.  
 *   Outputs
 *      detect  imaging_detector_table_def with results
 *      ierr    error code, 0=>OK
 *   Returns 0 if succeeds, else failed.
 **************************************************************** */
int fitDistort2 (imaging_detector_table_def *detect, 
		 int ndata, int ncoef, int iregF, int iregR,
		 char **files, int *ierr)
{
  int i, j, ifile, iter, icoef, mcoef, retCode = 1;
  int iXY, iRow, convgd, bestXY, bestCoef;
  float coef[2][MAXCOEF], tcoef[2][MAXCOEF], peakMax;
  float delta[2][MAXCOEF], peak[2][MAXCOEF];
  float tol, wx, test, norm, OF, lastOF;
  float OFCenter, OFPlus, OFMinus, deltaCoef;
  float tdelta[MAXCOEF] = {1.0, 0.01, 0.01, 0.0001, 0.0001, 0.0001};
  imaging_fdata_table_def *idata;
  imaging_fdata_row **rows=NULL, *work=NULL;
  fitsfile *inFile;

  /* sanity checks */
  if (*ierr!=0) return retCode; /* Previous Error condition? */
  if (ndata<1) return retCode;
  if (ncoef<1) return retCode;
  if (!files) return retCode;
  if (ncoef>MAXCOEF) ncoef = MAXCOEF;

  /* Create array for MIDI frames */
  rows = (imaging_fdata_row**)malloc(ndata*sizeof(imaging_fdata_row*));

  /* read data */
  for (ifile=0; ifile<ndata; ifile++) {
    /*    Open file */
    fits_open_file(&inFile, files[ifile], READONLY, ierr);
    if (*ierr !=0) {
      fits_report_error(stderr, *ierr);
      return retCode;
    }

    /*    Open table*/
    idata = open_imaging_fdata_table(inFile, ierr);
    if (*ierr !=0) {
      fits_report_error(stderr, *ierr);
      return retCode;
    }
    
    /*    Create table row */
    rows[ifile] = make_imaging_fdata_row(idata);
    /* Make work row if not yet done */
    if (!work) work = make_imaging_fdata_row(idata);
   
    /* Read first (only) row */
    iRow = 1;
    read_imaging_fdata_table(inFile, idata, rows[ifile], iRow, ierr);
    if (*ierr !=0) {
      fits_report_error(stderr, *ierr);
      return retCode;
    }

    /* Dedistort reference region */
    minrtsDedistortFrame (detect, iregR, rows[ifile], rows[ifile]);

    /*   Close file and table */
    close_imaging_fdata_table(inFile, idata, ierr);
    if (*ierr !=0) {
      fits_report_error(stderr, *ierr);
      return retCode;
    }
    /* blows up kill_imaging_fdata_table_def (idata);*/
    
  } /* end loop over files */

  /* initialize distortion from detect */
  for (i=0; i<MAXCOEF; i++) {coef[0][i] = 0.0; coef[1][i] = 0.0;}
  minrtsGetPoly  (detect, 0, iregF,  &mcoef, coef[0]);
  minrtsGetPoly  (detect, 1, iregF,  &mcoef, coef[1]);

  /********* Iteration loop ********/
  lastOF = 0.0;
  for (iter=1; iter<=50; iter++) {
    convgd = 0;  /* Can decide converged later */

    /* Find peak OF for each coefficient in range tdelta[coef]  */
    /* Get objective function at nominal value */
    minrtsSetPoly (detect, 0, iregF, ncoef, &coef[0][0]);
    minrtsSetPoly (detect, 1, iregF, ncoef, &coef[1][0]);
    OFCenter = ObjectFun (detect, iregF, iregR, ndata, rows, work);
    if (iter==1) lastOF = OFCenter;

    /* Find best performer */
    bestXY   = -1;
    bestCoef = -1;
    peakMax = -1.0;
    /* Loop over X, Y */
    for (iXY=0; iXY<=1; iXY++) {

      /* Loop over coefficients finding peak */
      for (icoef=0; icoef<MAXCOEF; icoef++) tcoef[iXY][icoef] = coef[iXY][icoef];
      for (icoef=0; icoef<ncoef; icoef++) {
	
	/* plus and minus deltaCoef */
	deltaCoef = tdelta[icoef];  /* test change in parameter by order */
	tcoef[iXY][icoef] = coef[iXY][icoef] + deltaCoef;
	minrtsSetPoly (detect, iXY, iregF, ncoef, &tcoef[iXY][0]);
	OFPlus = ObjectFun (detect, iregF, iregR, ndata, rows, work);
	tcoef[iXY][icoef] = coef[iXY][icoef] - deltaCoef;
	minrtsSetPoly (detect, iXY, iregF, ncoef, &tcoef[iXY][0]);
	OFMinus = ObjectFun (detect, iregF, iregR, ndata, rows, work);
	tcoef[iXY][icoef] = coef[iXY][icoef];

	/* debug
	fprintf (stdout, "iXY %d coef %d delta %f %f %f %f\n",
		 iXY, icoef, deltaCoef, OFMinus, OFCenter, OFPlus); */
	
	/* Find peak and offset in three determinations of OF */
	PolyFitPeak (OFMinus, OFCenter, OFPlus, deltaCoef, 
		     &delta[iXY][icoef], &peak[iXY][icoef]);

	/* Look for best */
	if (peak[iXY][icoef]>peakMax) {
	  peakMax = peak[iXY][icoef];
	  bestXY           =  iXY;
	  bestCoef         = icoef;
	}
      } /* end loop over coefficients */
    } /* end loop over X, Y */

    /* trial update solutions with damping to force motion towards convergence */
    wx = 2.0;
    while (wx>1.0e-10) { /* loop 'til improvement - but not forever */
      wx *= 0.5;

      /* get gradient norm */
      norm = 0.0;
      /* Convergence criteria, lower bar each attempt */
      tol = 0.5e-6 + iter*1.0e-5;

      /* Convergence test - Loop over X, Y */
      for (iXY=0; iXY<=1; iXY++) {
	for (i=0; i<ncoef; i++) {
	  test = tol * fabs(coef[iXY][i]);
	  /* Update only the best performer
	  if ((iXY==bestXY) && (i==bestCoef)) { */
	    tcoef[iXY][i] = coef[iXY][i] + wx * delta[iXY][i];  /* trial update */
	    convgd = convgd && (fabs(delta[iXY][i])<test); /* convergence test */
	    /*  } else {
		tcoef[iXY][i] = coef[iXY][i];
		}*/
	  norm += wx*wx*delta[iXY][i]*delta[iXY][i];
	}
      } /* end loop over X/Y */

      /* Determine Objective function and decrease changes until it improves */
      minrtsSetPoly (detect, 0, iregF, ncoef, &tcoef[0][0]);
      minrtsSetPoly (detect, 1, iregF, ncoef, &tcoef[1][0]);
      OF = ObjectFun (detect, iregF, iregR, ndata, rows, work);

      /* Force OF to increase, when they do break the trial loop */
      if (OF>lastOF) break;
    } /* end trial update loop */

    /* if no improved solution found declare it done */
    if (OF<=lastOF) convgd = 1;
    lastOF = OF;  /* save objective function */

    /* save updated coefficients */
    for (i=0; i<ncoef; i++) {
      coef[0][i] = tcoef[0][i];
      coef[1][i] = tcoef[1][i];
    }

    /* are we converged?  If so break from loop */
    if (convgd || ((iter>5) && (norm<1.0e-8))) break;
    /* debug */
       fprintf (stdout," iter %d OF %g wx %f norm %g  %g %g %g %g %g %g\n", 
       iter, OF, wx, norm, coef[1][0], coef[1][1], coef[1][2], 
		coef[1][3], coef[1][4], coef[1][5]);
       /*       fprintf (stdout,"             %g %g %g %g %g %g\n", 
		delta[1][0], delta[1][1], delta[1][2],
		peak[1][0], peak[1][1], peak[1][2]); */
  } /* end iteration loop */

  /* debug */
  fprintf (stdout," Number of iterations %d OF %f\n", iter, OF);

  /* save results on detect */
  minrtsSetPoly (detect, 0, iregF, ncoef, &coef[0][0]);
  minrtsSetPoly (detect, 1, iregF, ncoef, &coef[1][0]);

  /* Cleanup */
  kill_imaging_fdata_row (work);
  for (i=0; i<ndata; i++) kill_imaging_fdata_row (rows[i]);
  free (rows);

  retCode = 0;
  return retCode;
} /* end fitDistort2 */

/***************************************************************
 *  Regrid first imaging_fdata row in Files and write to outfile
 * Inputs:
 *    nFiles      Number of files in Files
 *    Files       Names of files to process
 *    outfileName Name of imaging_fdata_table to write
 * Outputs:
 *    ierr     error code, 0=>OK
 **************************************************************** */
void testRegrid (int nFiles, char** Files, char *outfileName, int *ierr)
{
   fitsfile *inFile=NULL, *outFile=NULL;
   imaging_detector_table_def *DetectorDef = NULL;
   imaging_fdata_table_def *idata=NULL, *odata=NULL;
   imaging_fdata_row *inData=NULL, *outData=NULL;
   int i, ifile, iRow, nreg, nx, ny, ireg;
   float corr[MAXREGION];

   if (*ierr!=0) return; /* Previous Error condition? */

   /*    Open output file */
   i = fits_open_file(&outFile, outfileName, READWRITE, ierr);
   if (*ierr !=0) {
     fprintf(stderr, "error %i opening file %s\n", *ierr, outfileName);
     fits_report_error(stderr, *ierr);
     return;
   }

   /*     Get detector section */
   DetectorDef = read_imaging_detector_table(outFile, ierr);
   if (*ierr !=0) {
     fprintf(stderr, "error %i opening detector table in %s\n", 
	     *ierr, outfileName);
     fits_report_error(stderr, *ierr);
     return;
   }
   
   odata = open_imaging_fdata_table(outFile, ierr);
   if (*ierr !=0) {
     fprintf(stderr, "error %i opening imaging data in file %s\n", 
            *ierr, outfileName);
     fits_report_error(stderr, *ierr);
     return;
   }

   /*    Create table row */
   outData = make_imaging_fdata_row(odata);

   /* Loop over input files */
   for (ifile=0; ifile<nFiles; ifile++) {
     /*    Open input file */
     i = fits_open_file(&inFile, Files[ifile], READONLY, ierr);
     if (*ierr !=0) {
       fprintf(stderr, "error %i opening file %s\n", *ierr, Files[ifile]);
       fits_report_error(stderr, *ierr);
       return;
     }
     
     /*    Open input table */
     idata = open_imaging_fdata_table(inFile, ierr);
     if (*ierr !=0) {
       fprintf(stderr, "error %i opening imaging data in file %s\n", 
	       *ierr, Files[ifile]);
       fits_report_error(stderr, *ierr);
       return;
     }
     
     /*    Create table row */
     inData = make_imaging_fdata_row(idata);
     
     /*    Read row */
     iRow = 1;
     i = read_imaging_fdata_table(inFile, idata, inData, iRow, ierr);
     if (*ierr !=0) {
       fprintf(stderr, "error %i reading imaging data line %i in file %s\n", 
	       *ierr, iRow, Files[ifile]);
       fits_report_error(stderr, *ierr);
       return;
     }

     /* compare before */
     corr[0] = CrossCorr (DetectorDef, 0, 2, inData, inData);
     corr[1] = CrossCorr (DetectorDef, 1, 2, inData, inData);
     corr[2] = CrossCorr (DetectorDef, 2, 2, inData, inData);
     corr[3] = CrossCorr (DetectorDef, 3, 2, inData, inData);
     fprintf (stdout, "File %s Before correlations %g %g %g %g\n",
	      Files[ifile], corr[0], corr[1], corr[2], corr[3]);
     
     /* regrid */
     minrtsDedistortFrame (DetectorDef, -1, inData, outData);
     
     /* compare after */
     corr[0] = CrossCorr (DetectorDef, 0, 2, outData, outData);
     corr[1] = CrossCorr (DetectorDef, 1, 2, outData, outData);
     corr[2] = CrossCorr (DetectorDef, 2, 2, outData, outData);
     corr[3] = CrossCorr (DetectorDef, 3, 2, outData, outData);
     fprintf (stdout, "   After correlations %g %g %g %g\n",
	      corr[0], corr[1], corr[2], corr[3]);
     
     /*    Write modified row */
     iRow = ifile+1;
     i = write_imaging_fdata_table(outFile, odata, outData, iRow, ierr);
     if (*ierr !=0) {
       fprintf(stderr, "error %i writing imaging data line %i in file %s\n", 
	      *ierr, iRow, outfileName);
       fits_report_error(stderr, *ierr);
       return;
     }
    /*   Close input file and table */
     i= close_imaging_fdata_table(inFile, idata, ierr);
   } /* end loop over input files */

   /*   Close output file and table */
   i= close_imaging_fdata_table(outFile, odata, ierr);

   /* Cleanup */
   kill_imaging_fdata_row(inData);
   kill_imaging_fdata_row(outData);

   return;
} /* end testRegrid */

/***************************************************************
 * Sum the pixel by pixel product of a region of two 
 * imaging_fdata_rows.
 * Inputs:
 *    detector IMAGING_DETECTOR structure describing input
 *    ireg1    Region number in rowin1
 *    ireg2    Region number in rowin2
 *    rowin1   first operand
 *    rowin2   second operand
 * Returns:
 *    sum of the products of pixels
 **************************************************************** */
float CrossCorr (imaging_detector_table_def *detector, 
		 int ireg1, int ireg2, 
		 imaging_fdata_row *rowin1,  imaging_fdata_row *rowin2)
{
    int npix, ipix;
    double sum = 0.0, sumA=0.0, sumB=0.0;
    float out;

    /* loop over pixels - compute correlation coefficient */
    npix = detector->naxis[ireg1][0] * detector->naxis[ireg1][1];
    for (ipix=0; ipix<npix; ipix++) {
      sum  += rowin1->data[ireg1][ipix]*rowin2->data[ireg2][ipix];
      sumA += rowin1->data[ireg1][ipix]*rowin1->data[ireg1][ipix];
      sumB += rowin2->data[ireg2][ipix]*rowin2->data[ireg2][ipix];
    }
    
    out = (float)(sum / sqrt(sumA * sumB));
    return out;
} /* end CrossCorr */

/***************************************************************
 * Calculate the objective function for final fitting of 
 * dedistortion polynimials.
 * Calculates the sum of the correlation (sum of pixel products)
 * of two specified regions in a set of imaging_fdata_rows.
 * Inputs:
 *    detect   IMAGING_DETECTOR structure describing input,
 *             used for the dedistortion polynomials
 *    iregT    Test region number.
 *    iregR    Reference region number, this region is expected 
 *             to have been dedistorted in rows prior to call.
 *    nrow     Number of imaging_fdata_row in rows.
 *    rows     Array of imaging_fdata_rows.
 *    work     imaging_fdata_row to use as scratch
 * Returns:
 *    Objective function values
 **************************************************************** */
float ObjectFun (imaging_detector_table_def *detect, 
		 int iregT, int iregR, int nrows,
		 imaging_fdata_row **rows, imaging_fdata_row *work)
{
  float out=0.0, *savep;
  int i;

  /* loop over rows */
  for (i=0; i<nrows; i++) {

    /* regrid regions */
    minrtsDedistortFrame (detect, iregT, rows[i], work);

    /* Switch pointer from data region iregR on rows[i] to work */
    savep = work->data[iregR];
    work->data[iregR] = rows[i]->data[iregR];

    /* Get cross correlation */
    out += CrossCorr (detect, iregT, iregR, work, work);
    work->data[iregR] = savep; /* restore pointer */
} /* end loop over rows */

  /* normalize by number of comparisons */
  out /= (float)nrows;

  return out;
} /* end ObjectFun */

/***************************************************************
 *   Estimate the location and value of a peak in a function from
 *   three equally spaced measurements.
 *   Fits 2nd order polynomial to three points (x=+/-delta, 0)
 *   and estimate location and value of peak.
 *   Inputs
 *     y1    Function value at x1 = -delta
 *     y2    Function value at x2 =  0
 *     y3    Function value at x3 =  delta
 *     delta Spacing between x values
 *   Outputs
 *     xmax  x of peak, clipped at +/- delta
 *     peak  estimated peak
 **************************************************************** */
void PolyFitPeak (float y1, float y2, float y3, float delta, 
			 float *xmax, float *peak)
{
  float a0, a1, a2, x;

  /* Polynomial coefficients */
  a0 = y2;
  a1 = 0.5 * (y3-y1);
  a2 = y3 - a0 - a1;

  /* max where derivative = 0 */
  x = -a1 / (2.0*a2);

  /* clip to range */
  if (x>1.0)  x =  1.0;
  if (x<-1.0) x = -1.0;

  /* peak value */
  *peak = a0 + a1*x + a2*x*x;

  /* Convert to true x units */
  *xmax = x * delta;

} /* end PolyFitPeak */
 
